#include "xml.h"

template<class T> int StrReplace(T*s,int slen,int buflen,int idel, int ldel, const T*sins,int lins);

int StrReplace<char>(char*s,int slen,int buflen,int idel, int ldel, const char*sins,int lins) {
  if (buflen<4) return 0;
  if (slen<0) slen=lstrlenA(s)+1;
  if ((unsigned)idel>=(unsigned)slen) return 0;
  if (lins<0) lins=lstrlenA(sins);
  int move=lins-ldel;	// amount of grow
  if ((unsigned)(slen+move)>(unsigned)buflen) return 0;
  MoveMemory(s+idel+lins,s+idel+ldel,slen-idel-ldel);
  CopyMemory(s+idel,sins,lins);
  return slen+move;
}

int StrReplace<WCHAR>(WCHAR*s,int slen,int buflen,int idel, int ldel, const WCHAR*sins,int lins) {
  if (buflen<4) return 0;
  if (slen<0) slen=lstrlenW(s)+1;
  if ((unsigned)idel>=(unsigned)slen) return 0;
  if (lins<0) lins=lstrlenW(sins);
  int move=lins-ldel;	// amount of grow
  if ((unsigned)(slen+move)>(unsigned)buflen) return 0;
  MoveMemory(s+idel+lins,s+idel+ldel,(slen-idel-ldel)*2);
  CopyMemory(s+idel,sins,lins*2);
  return slen+move;
}

void*Node::operator new(size_t sz) {
 return HeapAlloc(hHeap,HEAP_GENERATE_EXCEPTIONS|HEAP_NO_SERIALIZE|HEAP_ZERO_MEMORY,sz);
}

int Xml::compareText(const char*t1,int l1,const char*t2,int l2){
  if (l1!=l2) return -1;
  if (t1==t2) return 0;
  return memcmp(t1,t2,l1);
}


bool Xml::onShift(const Xml&,Node*n,void*param) {
  shift_t&sh=*(shift_t*)param;
  sh(n->scope);
  sh(n->name);
  sh(n->value);
  return true;
}

bool Xml::shift(unsigned from,int by) {
 if (from>len) return false;	// start index too large
 if (!by) return true;			// do nothing
 if (len+by<from) return false;	// shrink too much: New <len> would be less than <from>
 len+=by;
 if (by<0) {
  memmove(buf+from,buf+from-by,len-from);
  buf=(char*)HeapReAlloc(Node::hHeap,HEAP_NO_SERIALIZE,buf,len);
 }else{
  buf=(char*)HeapReAlloc(Node::hHeap,HEAP_NO_SERIALIZE,buf,len);
  memmove(buf+from+by,buf+from,len-from-by);
#ifdef _DEBUG
  memset(buf+from,'X',by);		// fill gap
#endif
 }
 shift_t sh={from,by};
 enumNodes(root,Node::type_t(-1),0,onShift,&sh);
 return true;
}

// el = Element node newly created
// i = position in buf (onto whitespace or ">" or "/>" or "?>")
// On return: i onto ">" or "/>" or "?>" or erraneous character
// attribute nodes added to el->sub
bool Xml::parseAttr(Node*el,unsigned&i) {
  while (i<len) {
    if (buf[i]<=' ') {++i; continue;}	// skip whitespace
    switch (buf[i]) {
      case '>': return true;
      case '?':
      case '/': return i<len-1 && buf[i+1]=='>';
    }
    Node*k=new Node(Node::Attr);
    k->scope.cpMin=k->name.cpMin=i;
    while (i<len) {
      char c=buf[i];
      if (c>' ' && c!='=' && c!='>' && c!='?') {++i; continue;}
      k->name.cpMax=i;
      if (c=='=') {
	char q=buf[++i];
	if (q!='"' && q!='\'') return false;	// must be in single or double quotes for XML, not for HTML
	k->value.cpMin=++i;
	for(;i<len && buf[i]!=q;i++);
	if (i==len) return false;	// end of input without closing quote
	k->value.cpMax=i++;
      }
      k->scope.cpMax=i;
      break;
    }
    el->add_child_back(k);
  }
  return false;				// end of input without ">"
}

int Xml::parse() {
 root->scope.cpMax=len;
 Node*cur=root;
 for (unsigned i=0; i<len;) {
  if (buf[i]<=' ') {++i; continue;}
  extern volatile bool wantStop;
  if (wantStop) return -1;
  if (buf[i]=='<') switch (buf[++i]) {
   case '/': {
    if (cur->type!=Node::Element) return i;
    unsigned k=++i;	// save start of string
    while (i<len && buf[i]>' ' && buf[i]!='>') ++i;	// locate end of tag
    if (k!=i	// Not only "</>"
    && compareText(cur->name,k,i)) return k;	// Error! Tag must be the same
    cur->value.cpMax=k-2;	// where innerHTML ends
    while (i<len && buf[i]!='>') ++i;	// locate end marker (Bug: skipping false attributes)
    if (i==len) return i;	// No end marker
    cur->scope.cpMax=++i;	// now end of scope available
    cur=cur->parent;	// ascend
   }break;
   case '?': {
    Node*n=new Node(Node::ProcessingInstruction);
    n->scope.cpMin=i-1;
    n->name.cpMin=++i;
    for(;i<len;i++) {
     char c=buf[i];
     if (c<=' ' || c=='>' || c=='/' || c=='?') break;
    }
    n->name.cpMax=i;
    if (!parseAttr(n,i)) return i;
    if (buf[i]!='?') return i;	// wrong termination
    n->scope.cpMax=(i+=2);
    cur->add_child_back(n);
   }break;
   case '!':
   if (!memcmp(buf+i+1,"--",2)) {
    Node*n=new Node(Node::Comment);
    n->scope.cpMin=i-1;		// auf "<!--"
    i+=2;
    if (i>len-4) return i;		// no space for " -->"
    n->value.cpMin=++i;
    for (i+=2; i<len; i++) if (buf[i-2]=='-' && buf[i-1]=='-' && buf[i]=='>') break;
    if (i==len) return i;		// missing termination
    n->value.cpMax=i-2;
    n->scope.cpMax=++i;
    cur->add_child_back(n);
    break;
   }
   if (!memcmp(buf+i+1,"DOCTYPE",7)) {
    Node*n=new Node(Node::DocumentType);
    n->scope.cpMin=i-1;
    for (i+=8; i<len; i++) if (buf[i]>' ') break;
    n->value.cpMin=i;
    for (; i<len; i++) if (buf[i]=='>') break;
    if (i==len) return i;		// Fehlendes Ende
    n->value.cpMax=i;
    n->scope.cpMax=++i;
    cur->add_child_back(n);
    break;
   }
   if (!memcmp(buf+i+1,"[CDATA[",6)) {
    Node*n=new Node(Node::CDATASection);
    n->scope.cpMin=i-1;
    for (i+=7; i<len; i++) if (buf[i]>' ') break;
    n->value.cpMin=i;
    for (i+=2; i<len; i++) if (buf[i-2]==']' && buf[i-1]==']' && buf[i]=='>') break;
    if (i==len) return i;		// Fehlendes Ende
    n->value.cpMax=i;
    n->scope.cpMax=++i;
    cur->add_child_back(n);
    break;
   }//fall-through
   default:{
    Node*n=new Node(Node::Element);
    n->scope.cpMin=i-1;
    n->name.cpMin=i;
    char c;
    for(;i<len;i++) if ((c=buf[i])<=' ' || c=='>' || c=='/' || c=='?') break;
    n->name.cpMax=i;
    if (!parseAttr(n,i)) return i;
    c=buf[i];
    if (c=='?') return i;		// not allowed end marker
    cur->add_child_back(n);
    if (c=='/') {
     n->scope.cpMax=(i+=2);
     break;
    }
    n->value.cpMin=++i;	// where innerHTML starts
    cur=n;	// descend
   }
  }else{
   if (cur->type!=Node::Element) return i;	// root-level text not allowed
   Node*n=new Node(Node::Text);
   n->scope.cpMin=i;
   while (i<len && (buf[i]<=' ' || buf[i]=='<')) ++i;
   n->value.cpMin=i;
   while (i<len && buf[i]!='<') ++i;
   n->scope.cpMax=i;
   int j=i;
   while (buf[--j]<=' ');
   n->value.cpMax=max(j+1,n->value.cpMin);
   cur->add_child_back(n);
  }
 }
 if (cur->type!=Node::root) return i;	// Not at root at end of input
 return 0;	// okay
}

Node*Xml::findNode(const Node*r,Node::type_t t,const char*tag) const{
  Node*ret=0;
  enumNodes(const_cast<Node*>(r),t,tag,stop_and_save,&ret);
  return ret;
}

bool Xml::enumNodes(Node*r,Node::type_t t,const char*tag,bool(*cb)(const Xml&,Node*,void*),void*param) const{
  if (!r) return true;
  Node*n=r;
  do{
    if ((t==-1 || n->type==t) && (!tag || !compareText(n->name,tag)) && !cb(*this,n,param)) return false;
    if (!enumNodes(n->sub,t,tag,cb,param)) return false;	// recurse through sub-tree
    n=n->next;
  }while (n!=r);	// stop when all siblings are checked
  return true;
}

Node*Xml::findChildNode(const Node*p,Node::type_t t,const char*tag,int tlen) const{
  if (tlen<0) tlen=strlen(tag);
  Node*n0=p->sub,*n=n0;
  if (!n0) return 0;
  do{
    if ((t==-1 || n->type==t) && (!compareText(n->name,tag,tlen))) return n;
    n=n->next;
  }while (n!=n0);	// stop when all siblings are checked
  return 0;
}

HANDLE Node::hHeap;

void Xml::show(HWND hEdit) const{
  int lw=MultiByteToWideChar(CP_UTF8,0,buf,len,0,0);
  TCHAR*sw=(TCHAR*)_alloca((lw+1)*sizeof(TCHAR));
  MultiByteToWideChar(CP_UTF8,0,buf,len,sw,lw);
  sw[lw]=0;
  Edit_SetModify(hEdit,true);
  SetWindowText(hEdit,sw);	// EN_CHANGE enabled dann den Apply-Menpunkt - denkste!! Nicht bei RichEdit!
  Edit_SetModify(hEdit,true);
}

bool Xml::deleteNode(Node*n) {
  if (n->type==Node::root) return false;
  int a=n->scope.cpMin;
  int e=n->scope.cpMax;
  if (buf[e]==' ') ++e;
  if (buf[e]=='\r') ++e;
  if (buf[e]=='\n') ++e;
  Node*prv=n->prev,*nxt=n->next;
  prv->next=nxt;	// may write to same node when there is only one sibling
  nxt->prev=prv;
  if (n->parent->sub==n) n->parent->sub=nxt!=n?nxt:0;
  delete n;
  return shift(a,a-e);
}

//Leeres Elternelement der Form <tag/>\n oder <tag attr="value"/>\n aufteilen in <tag>\n</tag>\n bzw. <tag attr="value">\n</tag>\n.
//Dann erst kann das Kindelement (auer Attr) eingefgt werden.
void Xml::splitElement(Node*el) {
  if (el->type!=Node::Element) return;
  if (el->value.cpMin) return;	// already split (value is innerHTML)
  int a=el->scope.cpMax-2;	// onto "/>"
  int b=el->name.cpMax-el->name.cpMin;
  if (!shift(a,b+3)) return;	// make gap, modifying el->scope.cpMax
  _snprintf(buf+a,b+5,"> </%.*s>",b,buf+el->name.cpMin);
  el->value.cpMin=a+1;		// onto " "
  el->value.cpMax=a+2;		// onto "</tag>"
}

void Xml::unsplitElement(Node*el) {
  if (el->type!=Node::Element) return;
  if (!el->value.cpMin) return;	// already unsplit
  Node*q,*p=el->sub;
  if (q=p) do{
    if (q->type!=Node::Attr) return;	// has non-attribute child
  }while((q=q->next)!=p);		// until end of list
  int a=el->value.cpMin+1;	// one character after ">": Let space for "/>"
  int b=el->scope.cpMax-a;
  shift(a,-b);
  _snprintf(buf+a-2,2,"/>");
  el->value.cpMax=el->value.cpMin=0;	// mark as unsplit form
}

Node*Xml::addElement(Node*p,const char*tag,int tlen,Node*before) {
  if (!p) p=root;
  if (!(p->type==Node::Element || p->type==Node::root)) return 0;	// parent must be element or root
  if (before) {
    if (before->parent!=p) return 0;	// <before> must be direct child of <p>
    if (before->type==Node::Attr) return 0;	// cannot be an attribute
  }
  if (tlen<0) tlen=strlen(tag);
  splitElement(p);
  int a=before?before->scope.cpMin:p->sub&&p->sub->prev->type!=Node::Attr?p->sub->prev->scope.cpMax:p->value.cpMax;
  if (!a) return 0;
  int b=tlen+5;
  if (!shift(a,b)) return 0;
  _snprintf(buf+a,b,"\n<%.*s/>\n",tlen,tag);	// create an unsplit element in a new line
  Node*n=new Node(Node::Element);
  n->scope.cpMin=a+1;
  n->scope.cpMax=a+b-2;
  n->name.cpMin=a+2;
  n->name.cpMax=a+2+tlen;
  if (before) before->add_sibling_back(n);
  else p->add_child_back(n);
  return n;  
}

Node*Xml::setElement(Node*p,const char*tag,bool*created) {
  if (created) *created=false;
  Node*n=findChildNode(p,Node::Element,tag);	// wenn's da ist, liefern
  if (n) return n;
  if (p==root) {
    Node*pi=addPI("xml",3,p->sub);	// am Anfang
    addAttr(pi,"version","1.0");
    addAttr(pi,"encoding","UTF-8");
    addAttr(pi,"standalone","yes");
  }
  if (created) *created=true;
  return addElement(p,tag);			// wenn nicht, erzeugen und (innen) anhngen
}

Node*Xml::addPI(const char*tag,int tlen,Node*before) {
  if (before && before->parent!=root) return 0;	// <before> must be direct child of root
  if (tlen<0) tlen=strlen(tag);
  int a=before?before->scope.cpMin:0;
  int b=tlen+6;
  if (!shift(a,b)) return 0;
  _snprintf(buf+a,b,"<?%.*s ?>\n",tlen,tag);	// create a processing instruction in a new line
  Node*n=new Node(Node::ProcessingInstruction);
  n->scope.cpMin=a;
  n->scope.cpMax=a+b-1;
  n->name.cpMin=a+2;
  n->name.cpMax=a+2+tlen;
  if (before) before->add_sibling_back(n);
  else root->add_child_back(n);
  return n;  
}

Node*Xml::addText(Node*p,const char*txt,int tlen,Node*before) {
  if (!p) return 0;
  if (p->type!=Node::Element) return 0;
  if (before) {
    if (before->parent!=p) return 0;	// <before> must be direct child of <p>
    if (before->type==Node::Attr) return 0;	// cannot be an attribute
  }
// TODO: Check whether insertion point is not surrounded by text nodes, then modify text node instead
  splitElement(p);
  int a=before?before->scope.cpMin:p->sub?p->sub->prev->scope.cpMax:p->value.cpMin;
  if (!a) return 0;	// something wrong
  if (!shift(a,tlen)) return 0;
  _snprintf(buf+a,tlen,"%.*s",tlen,txt);	// dasselbe wie memcpy(buf+a,txt,tlen)
  Node*n=new Node(Node::Text);
  n->scope.cpMin=n->value.cpMin=a;
  n->scope.cpMax=n->value.cpMax=a+tlen;
  if (before) before->add_sibling_back(n);
  else p->add_child_back(n);
  return n;
}

Node*Xml::addAttr(Node*el,const char*tag,int tlen,const char*val,int vlen,Node*before) {
  if (!el) return 0;
  if (el->type!=Node::Element) return 0;
  if (before) {
    if (before->parent!=el) return 0;
    if (before->type!=Node::Attr) return 0;
  }
  Node*lastattr=0;
  if (!before) {
    Node*q,*p=el->sub;
    if (q=p) do{
      if (q->type!=Node::Attr) break;
      lastattr=q;
    }while((q=q->next)!=p);
  }
  int a=before?before->scope.cpMin-1:lastattr?lastattr->scope.cpMax:el->name.cpMax;
  if (!a) return 0;
  if (tlen<0) tlen=strlen(tag);
  if (vlen<0) vlen=strlen(val);
  int b=tlen+vlen+4;
  if (!shift(a,b)) return 0;
  _snprintf(buf+a,b," %.*s=\"%.*s\"",tlen,tag,vlen,val);
  Node*n=new Node(Node::Attr);
  n->scope.cpMin=n->name.cpMin=a+1;
  n->name.cpMax=a+1+tlen;
  n->value.cpMin=a+1+tlen+2;
  n->value.cpMax=a+b-1;
  n->scope.cpMax=a+b;
  if (before) before->add_sibling_back(n);
  else if (lastattr) lastattr->add_sibling_front(n);
  else el->add_child_back(n);
  return n;
}

