#define UNICODE
#define _UNICODE
#include <windows.h>
#include <windowsx.h>	// Edit_SetModify
#include <malloc.h>	// _alloca
#include <stdio.h>	// _snprintf
//#include <richedit.h>	// CHARRANGE

template<class T> int StrReplace(T*s,int slen,int buflen,int idel, int ldel, const T*sins,int lins);

int StrReplace<char>(char*s,int slen,int buflen,int idel, int ldel, const char*sins,int lins);
int StrReplace<WCHAR>(WCHAR*s,int slen,int buflen,int idel, int ldel, const WCHAR*sins,int lins);

struct CHARRANGE{
 long cpMin,cpMax;
 long length() const {return cpMax-cpMin;}
 bool defined() const {return !!cpMax;}
};

// <Node> macht sich nicht die Mhe, irgendwelche Strings zu speichern, sondern ist lightweight
// mit fester Gre (8Int und 4Zeiger, macht 32+16=48 Byte (32 Bit), 32+32=64 Byte (64 Bit))
struct Node{
 static HANDLE hHeap;
 enum type_t{		//scope		sub			name	value
  root,			//0..len	PI,Element,Comment	0	0
  Element,		//<tag..>	Attr;Element,Text,Comm.	tag	0 if collapsed, between > and </ when split
  Attr,			//tag="value"	0			tag	value
  Text,			//text		Entity			0	text without surrounding whitespace
  CDATASection,		//<![CDATA[var]]> 0			0	var
  EntityReference,	//?
  Entity,		//&tag;		0			tag	0
  ProcessingInstruction,//<?tag ?>	Attr			tag	between tag and ?>
  Comment,		//<!-- -->	0			0	between <!-- and -->
  Document,		//HTML only
  DocumentType,		//<!DOCTYPE >
  DocumentFragment,	//HTML only
  Notation		//?
 }type;
 int depth;
 CHARRANGE scope,name,value;
 Node*next,*prev,*parent,*sub;
 static void*operator new(size_t sz);
 static void operator delete(void*) {}	// simply don't delete
 Node(type_t t=root):type(t),depth(0) {next=prev=this;}
//  Node(type_t t,int a, int e):type(t),depth(0) {next=prev=this; scope.cpMin=a; scope.cpMax=e;}
//  int build(int);
//  Node*&lastsub() {if (sub) {Node**p=&sub; while ((*p)->next) p=&((*p)->next); return *p;} return sub;}
 void add_child_back(Node*n) {if (sub) sub->add_sibling_back(n); else{sub=n;n->parent=this;n->depth=depth+1;}}
//  Node*last() {Node*p=this; while (p->next) p=p->next; return p;}
 void add_sibling_back(Node*n) {n->parent=parent; n->next=this; prev->next=n; n->prev=prev; prev=n;n->depth=depth;}
 void add_sibling_front(Node*n) {add_sibling_back(n); n->parent->sub=n;}
 bool isChild(Node*n) const {return n->parent==this;}
 bool isLastChild() const {return !next || next==parent->sub;}
 bool isFirstChild() const {return this==parent->sub;}
};

struct Xml {
 Node*root;
 char*buf;	// UTF-8
 unsigned len;
// Unsauber: Wie bekommt man ein operator new mit hHeap in Xml hin??
 int parse();	// linearisierte Version!
 Node*addAttr(Node*p,const char*tag,int tlen,const char*val,int vlen=-1,Node*before=0);
 Node*addAttr(Node*p,const char*tag,const char*val,Node*before=0) {return addAttr(p,tag,-1,val,-1,before);}
 Node*addElement(Node*p,const char*tag,int tlen=-1,Node*before=0/*,Node::type_t t=Node::Element*/);
 Node*setElement(Node*p,const char*tag,bool* =0);	// Kind von p finden oder anhngen
 Node*addPI(const char*tag,int tlen=-1,Node*before=0);	// always to root node
 Node*addText(Node*p,const char*txt,int tlen=-1,Node*before=0);
 void splitElement(Node*el);	// prepare for text and subelement insertion
 void unsplitElement(Node*el);	// after text and subelement deletion
 bool deleteNode(Node*n);
 void show(HWND) const;
 bool shift(unsigned from,int by);	// Text ab Index um Delta (einfgen, lschen) sowie Nodes anpassen
 bool parseAttr(Node*el,unsigned&i);
 int compareText(const CHARRANGE&t1,const CHARRANGE&t2) const {return compareText(t1,t2.cpMin,t2.cpMax);}
 int compareText(const CHARRANGE&t1,int a,int e) const {return compareText(t1,buf+a,e-a);}
 int compareText(const CHARRANGE&t1,const char*t2) const {return compareText(t1,t2,strlen(t2));}
 int compareText(const CHARRANGE&t1,const char*t2,int len) const {return compareText(buf+t1.cpMin,t1.cpMax-t1.cpMin,t2,len);}
 static int compareText(const char*t1,int l1,const char*t2,int l2);
 Node*findElement(const Node*r,const char*tag) const {return findNode(r,Node::Element,tag);}
 Node*findNode(const Node*r,Node::type_t t,const char*tag=0) const;
 bool enumNodes(Node*r,Node::type_t t,const char*tag,bool(*)(const Xml&,Node*,void*),void*) const;
 Node*findChildNode(const Node*p,Node::type_t t,const char*tag,int tlen=-1) const;
private:
 static bool stop_and_save(const Xml&,Node*n,void*param) {*(Node**)param=n; return false;}
 struct shift_t{
  int from,by;
  operator()(CHARRANGE&r)	{if (r.cpMin>=from) r.cpMin+=by; if (r.cpMax>from) r.cpMax+=by;}
 };
 static bool onShift(const Xml&,Node*n,void*param);
};
